









local abs = math.abs
local floor = math.floor
local max = math.max
local min = math.min
local sort = table.sort

local vehicle_types =
{
  ["locomotive"] = true,
  ["cargo-wagon"] = true,
  ["car"] = true,
}
local train_types =
{
  ["locomotive"] = true,
  ["cargo-wagon"] = true,
}
local default_masses =
{
  ["locomotive"] = 200000,
  ["cargo-wagon"] = 100000,
  ["car"] = 3000,
}

-- Local function prototypes
local cache_equipment
local cars_tick
local charge_tick
local disable_train_braking
local disable_vehicle_braking
local enable_train_braking
local enable_vehicle_braking
local fuel_tick
local handle_train_state_change
local on_entity_removed
local on_train_removed
local read_equipment
local recover_energy
local update_equipment

-- Table utils
-------------------------------------------------------------------------------
-- These utilities assume the table is an array and each value is unique

local function insert_unique_f(tbl, x, f)
  for k, v in pairs(tbl) do
    if(f(v)) then
      return false
    end
  end
  tbl[#tbl + 1] = x
  return true
end

local function erase_f(tbl, f)
  for k, v in pairs(tbl) do
    if(f(v)) then
      tbl[k] = nil
      return k
    end
  end
  return nil
end

-- Local caches
-------------------------------------------------------------------------------

-- Dictionary of only those vehicles that have a transformer
-- [unit-number] = transformer index
local transformer_for_unit = { }
-- Dictionary of only those vehicles that have a regenerative brake
-- [unit-number] = brake index
local regen_brake_for_unit = { }

function rebuild_caches()
  equipment_cache = { }
  transformer_for_unit = { }
  regen_brake_for_unit = { }
  
  for unit, entity in pairs(global.vehicles) do
    local grid = entity.grid
    cache_equipment(unit, read_equipment(unit, grid))
  end
end

function validate_prototypes()
  -- Stop doing anything involving a no longer present prototypes
  local equipment = game.equipment_prototypes
  local entities = game.entity_prototypes
  for tbl, exists in pairs{[global.transformers] = equipment,
                           [global.brakes] = equipment,
                           [global.car_masses] = entities} do
    for name in pairs(tbl) do
      if(not exists[name]) then
        tbl[name] = nil
      end
    end
  end
end

function validate_entities()
  -- If any units were deleted stop processing them
  for unit, entity in pairs(global.vehicles) do
    if(not entity.valid or not entity.grid) then
      global.vehicles[unit] = nil
    else
      -- In case equipment was deleted
      update_equipment(unit, entity.grid)
    end
  end
  -- Stop regeneration on deleted vehicles and trains
  for _, tbl in pairs{global.braking_vehicles, global.braking_trains} do
    for i, data in pairs(tbl) do
      if(not tbl[i][1].valid) then
        tbl[i] = nil
      end
    end
  end
end

-- Equipment handling
-------------------------------------------------------------------------------

read_equipment = function(unit, grid)
  local equipment = grid.equipment
  local known_transformers = global.transformers
  local known_brakes = global.brakes
  local transformers = { }
  local brakes = { }
  for i = 1, #equipment do
    local item = equipment[i]
    if(known_transformers[item.name] and item.prototype.energy_source) then
      transformers[#transformers + 1] = i
    elseif(known_brakes[item.name] and item.prototype.energy_source) then
      brakes[#brakes + 1] = i
    end
  end
  sort(transformers, function(a, b) return equipment[a].prototype.energy_source.input_flow_limit >
                                           equipment[b].prototype.energy_source.input_flow_limit end)
  sort(brakes, function(a, b) return known_brakes[equipment[a]][2] >
                                     known_brakes[equipment[b]][2] end)
  return transformers, brakes
end

cache_equipment = function(unit, transformers, brake)
  transformer_for_unit[unit] = transformers[1]
  regen_brake_for_unit[unit] = brake[1]
end

function update_equipment(unit, grid)
  local transformers, brakes = read_equipment(unit, grid)
  
  local entity = global.vehicles[unit]
  local had_brake = regen_brake_for_unit[unit] ~= nil
  cache_equipment(unit, transformers, brakes)

  -- If the brake was removed disable regen braking
  if(not brakes[1] and had_brake) then
    if(not train_types[entity.type]) then
      disable_vehicle_braking(entity)
    end
  elseif(not had_brake and brakes[1]) then
    if(entity.type == "car") then
      enable_vehicle_braking(entity)
    end
  end
end

-- on_tick actions
-------------------------------------------------------------------------------

fuel_tick = function()
  local vehicles = global.vehicles
  for unit, equipment in pairs(transformer_for_unit) do
    local entity = vehicles[unit]
    local transformer = entity.grid.equipment[equipment]
    local available_energy = transformer.energy
    local current_energy = entity.energy
    entity.energy = current_energy + available_energy
    -- The burner energy buffer is kinda weird so we have to check how much energy we actually inserted
    transformer.energy = available_energy - (entity.energy - current_energy)
  end
end

recover_energy = function()
  local brakes = global.brakes
  for _, data in pairs(global.braking_trains) do
    local train = data[1]
    local previous_speed = data[2]
    local mass = data[3]
    local speed = train.speed * 60 -- convert m/tick to m/s
    local abs_speed = abs(speed)
    local difference = previous_speed - abs_speed
    if(difference > 0) then
      local previous_energy = 0.5 * mass * previous_speed * previous_speed
      local current_energy =  0.5 * mass * abs_speed * abs_speed
      local gained = previous_energy - current_energy -- TODO: consider friction
      local locos
      -- TODO: support for bidirectional locomotives?
      if(speed > 0) then
        locos = train.locomotives.front_movers
      else
        locos = train.locomotives.back_movers
      end
      if(#locos > 0) then
        gained = gained / #locos
        for j = 1, #locos do
          local loco = locos[j]
          local brake = regen_brake_for_unit[loco.unit_number]
          if(brake) then
            local item = loco.grid.equipment[brake]
            item.energy = item.energy + gained * brakes[item.name][2]
          end
        end
      end
    end
    data[2] = abs_speed
  end
  for unit, data in pairs(global.braking_vehicles) do
    local entity = data[1]
    local brake = regen_brake_for_unit[unit]
    if(brake) then
      local previous_speed = data[2]
      local speed = entity.speed * 60 -- convert m/tick to m/s
      local abs_speed = abs(speed)
      local difference = previous_speed - abs_speed
      if(difference > 0) then
        local previous_energy = 0.5 * mass * previous_speed * previous_speed
        local current_energy =  0.5 * mass * abs_speed * abs_speed
        local gained = previous_energy - current_energy -- TODO: consider friction
        local item = entity.grid.equipment[brake]
        item.energy = item.energy + gained * brakes[item.name][2]
      end
    end
  end
end

-- Braking
-------------------------------------------------------------------------------

enable_vehicle_braking = function(entity)
  local data = global.vehicle_data[entity.name]
  local mass = data and data[2] or default_masses[entity.type]
  if(mass) then
    global.braking_vehicles[entity.unit_number] =
    {
      [1] = entity,
      [3] = mass,
      [2] = entity.speed * 60, -- convert m/tick to m/s
    }
  end
end

disable_vehicle_braking = function(entity)
  global.braking_vehicles[entity.unit_number] = nil
end

enable_train_braking = function(train)
  -- We currently assume fixed masses for locomotive = 200 tons and cargo wagons = 100 tons.
  -- Ideally we could incorporate the prototype's weight property but it isn't available and has an unknown unit.
  local mass = 0
  local carriages = train.carriages
  local vehicle_data = global.vehicle_data
  for i = 1, #carriages do
    local data = vehicle_data[carriages[i].name]
    mass = mass + (data and data[2] or default_masses[carriages[i].type])
  end
  insert_unique_f(global.braking_trains,
                  {
                    [1] = train,
                    [2] = abs(train.speed * 60), -- convert m/tick to m/s
                    [3] = mass,
                  },
                  function(data) return data[1] == train end)
end

disable_train_braking = function(train)
  erase_f(global.braking_trains, function(data) return data[1] == train end)
end

-- Train specific
-------------------------------------------------------------------------------

handle_train_state_change = function(train)
  local state = train.state
  local braking = state == defines.train_state.path_lost or
                  state == defines.train_state.arrive_signal or
                  state == defines.train_state.arrive_station or
                  state == defines.train_state.manual_control_stop or
                  state == defines.train_state.stop_for_auto_control
  if(braking) then
    enable_train_braking(train)
  else
    disable_train_braking(train)
  end
end

-- Entity management
-------------------------------------------------------------------------------

on_train_removed = function(train)
  disable_train_braking(train)
end

on_entity_removed = function(entity)
  if(train_types[entity.type]) then
    local train = entity.train
    if(train and #train.carriages <= 1) then
      on_train_removed(train)
    end
  end
  if(vehicle_types[entity.type]) then
    local unit = entity.unit_number
    global.vehicles[unit] = nil
    global.braking_vehicles[unit] = nil
    transformer_for_unit[unit] = nil
    regen_brake_for_unit[unit] = nil
  end
end

-- Event entry points
-------------------------------------------------------------------------------

function on_built_entity(event)
  local entity = event.created_entity
  if(vehicle_types[entity.type] and entity.grid) then
    global.vehicles[entity.unit_number] = entity
    update_equipment(entity.unit_number, entity.grid)
  end
end

function on_entity_died(event)
  on_entity_removed(event.entity)
end

function on_player_placed_equipment(event)
  for unit, entity in pairs(global.vehicles) do
    if(entity.grid == event.grid) then
      update_equipment(unit, event.grid)
      break
    end
  end
end

function on_player_removed_equipment(event)
  for unit, entity in pairs(global.vehicles) do
    if(entity.grid == event.grid) then
      update_equipment(unit, event.grid)
      break
    end
  end
end

function on_preplayer_mined_item(event)
  on_entity_removed(event.entity)
end

function on_robot_pre_mined(event)
  on_entity_removed(event.entity)
end

function on_tick(event)
  fuel_tick()
  recover_energy()
end

function on_train_changed_state(event)
  handle_train_state_change(event.train)
end

-- Remote interface
-------------------------------------------------------------------------------

function register_transformer(data)
  assert(type(data.name) == "string", "'name' must be a string")
  local prototype = game.equipment_prototypes[data.name]
  assert(prototype, string.format("%s is not a valid equipment prototype", data.name))
  assert(prototype.energy_source, string.format("%s has no energy_source", data.name))
  global.transformers[data.name] =
  {
    [1] = data.name,
    [2] = prototype.energy_source.input_flow_limit,
  }
end

function register_brake(data)
  assert(type(data.name) == "string", "'name' must be a string")
  assert(game.equipment_prototypes[data.name], string.format("%s is not a valid equipment prototype", data.name))
  assert(type(data.efficiency) == "number" and data.efficiency >= 0 and data.efficiency <= 1,
         "brake efficiency must be a number in the range [0,1]")
  global.brakes[data.name] =
  {
    [1] = data.name,
    [2] = data.efficiency,
  }
end

function set_vehicle_data(data)
  assert(type(data.name) == "string", "'name' must be a string")
  assert(game.entity_prototypes[data.name], string.format("%s is not a valid entity prototype", data.name))
  assert(vehicle_types[game.entity_prototypes[data.name].type],
         string.format("%s is not a valid prototype ('car', 'locomotive' or 'cargo-wagon')", data.name))
  assert(type(data.mass) == "number" and data.mass > 0, "'mass' must be positive")
  global.vehicle_data[data.name] =
  {
    [1] = data.name,
    [2] = data.mass,
  }
end

function get_vehicle_data(name)
  local data = global.vehicle_data[name]
  if(data) then
    return {
      name = data[1],
      mass = data[2],
    }
  else
    return nil
  end
end
